/*
 * CyberPro 2000 Video Capture driver for the Rebel.com NetWinder
 *
 * (C) 1999 Russell King
 *
 *  Re-written from Rebel.com's vidcap driver
 *
 * Note: the maximum useful capture resolution is:
 *   625-line UK: 700x572
 *   525-line US: ?
 */
#include <linux/module.h>
#include <linux/videodev.h>
#include <linux/video_decoder.h>
#include <linux/mm.h>
#include <linux/i2c-old.h>
#include <linux/spinlock.h>
#include <linux/malloc.h>
#include <linux/delay.h>
#include <linux/sched.h>
#include <linux/init.h>

#include <asm/io.h>
#include <asm/irq.h>
#include <asm/pgtable.h>
#include <asm/hardware.h>

/*
 * virtual address of CyberPro registers
 */
static unsigned char *CyberRegs;
static char *fb_base;
static unsigned int fb_size;

#include "../video/cyber2000fb.h"

#undef USE_IRQ

#define BUFFER_ORDER	8
#define NR_FRAMES	2

struct cyberpro_grab_info {
	struct video_device *dev;

	int		buf_set:1;		/* VIDIOCSFBUF has been issued	*/
	int		win_set:1;		/* VIDIOCSWIN has been issued 	*/
	int		cap_active:1;		/* capture is active		*/
	int		ovl_active:1;		/* overlay is active		*/
	int		stream_8bpp:1;		/* stream from decoder is 8bpp	*/
	int		mmaped:1;		/* buffer is mmap()d		*/
	int		unused:26;

	int		users;			/* number of users		*/
	int		cap_width;		/* capture width		*/
	int		cap_height;		/* capture height		*/
	unsigned long	cap_mem_offset;		/* capture framebuffer offset	*/
	void *		buffer;			/* kernel capture buffer	*/
	int		frame_idx;		/* currently grabbing frame	*/

	struct video_capability	cap;		/* capabilities			*/
	struct video_channel	chan;		/* current channel settings	*/
	struct video_picture	pic;		/* current picture settings	*/
	struct video_window	win;		/* overlay window		*/
	struct video_buffer	buf;		/* display parameters		*/

	struct {
		unsigned int width;
		unsigned int height;
		unsigned int status;
#define UNUSED		0
#define GRABBING	1
#define DONE		2
	} frame[NR_FRAMES];

	wait_queue_head_t waitq;

	/*
	 * hw regs
	 */
	unsigned char	cap_mode1;
	unsigned char	cap_miscctl;
	unsigned char	disp_ctl1;
	unsigned char	vid_fmt;
	unsigned char	vfac1;
	unsigned char	vfac2;
	unsigned char	vfac3;
};

static void cyberpro_interrupt(int nr, void *dev_id, struct pt_regs *regs);

/*
 * On the NetWinder, we can put the SAA7111 to sleep.  We try
 * to keep it asleep as much as possible.  This generally means
 * whenever the capture device is not open.
 */
static void
decoder_sleep(int sleep)
{
#ifdef CONFIG_ARCH_NETWINDER
	extern spinlock_t gpio_lock;

	spin_lock_irq(&gpio_lock);
	cpld_modify(CPLD_7111_DISABLE, sleep ? CPLD_7111_DISABLE : 0);
	spin_unlock_irq(&gpio_lock);

	if (!sleep) {
		/*
		 * wait for device to wake up
		 */
		set_current_state(TASK_UNINTERRUPTIBLE);
		schedule_timeout(HZ / 50);
	}
#endif
}

static void cyberpro_grphw8(int reg, unsigned int val)
{
	cyber2000_grphw(reg, val);
}

static unsigned char cyberpro_grphr8(int reg)
{
	cyber2000_outb(reg, 0x3ce);
	return cyber2000_inb(0x3cf);
}

static void cyberpro_grphw16(int reg, unsigned int val)
{
	cyber2000_grphw(reg, val);
	cyber2000_grphw(reg + 1, val >> 8);
}

static void cyberpro_grphw24(int reg, unsigned int val)
{
	cyber2000_grphw(reg, val);
	cyber2000_grphw(reg + 1, val >> 8);
	cyber2000_grphw(reg + 2, val >> 16);
}

#define I2C_DELAY	100

static void
cyber2000fb_i2c_setlines(struct i2c_bus *bus, int ctrl, int data)
{
	unsigned long flags;
	int v;

	v = (ctrl ? 0x10 : 0x00) | (data ? 0x40 : 0x00);

	save_flags(flags);
	cli();

	cyber2000_grphw(0xb6, v);

	restore_flags(flags);

	udelay(I2C_DELAY);
}

static int
cyber2000fb_i2c_getdataline(struct i2c_bus *bus)
{
	unsigned long flags;
	int v;

	save_flags(flags);
	cli();

	cyber2000_outb(0xb6, 0x3ce);
	v = cyber2000_inb(0x3cf);

	restore_flags(flags);

	return v & 0x80 ? 1 : 0;
}

static void
cyber2000fb_i2c_null(struct i2c_bus *bus, int id)
{
}

static struct i2c_bus cyberpro_bus = {
	"",
	I2C_BUSID_CYBER2000,
	NULL,

	SPIN_LOCK_UNLOCKED,

	cyber2000fb_i2c_null,
	cyber2000fb_i2c_null,

	cyber2000fb_i2c_setlines,
	cyber2000fb_i2c_getdataline,
	NULL,
	NULL
};

static void reset_seq(void)
{
	unsigned char ext_mem_ctl;

	cyber2000_outb(0x70, 0x3ce);
	ext_mem_ctl = cyber2000_inb(0x3cf);

	cyber2000_outb(ext_mem_ctl | 0x80, 0x3cf);
	cyber2000_outb(ext_mem_ctl, 0x3cf);
}

static void
cyberpro_grabber_capture(struct cyberpro_grab_info *dp, int on)
{
	unsigned char old;

	/*
	 * freeze capture asynchronously
	 *  (FIXME: do we need to set VFAC_CTL1_FREEZE_CAPTURE_SYNC?)
	 */
	cyberpro_grphw8(VFAC_CTL1, dp->vfac1 |
				VFAC_CTL1_FREEZE_CAPTURE|
				VFAC_CTL1_FREEZE_CAPTURE_SYNC);

	if (!(cyberpro_grphr8(0xe5) & 0x02))
		printk(KERN_WARNING
			"%s: video capture freeze unsuccessful\n",
			dp->dev->name);

	/*
	 * Reset the memory sequencer
	 */
	reset_seq();

	old = cyberpro_grphr8(0xb5);

	if (on) {
		dp->vfac1 |= 1;

		cyberpro_grphw8(0xb5, old | 0x01);

		cyberpro_grphw8(VFAC_CTL1, dp->vfac1 |
				VFAC_CTL1_FREEZE_CAPTURE|
				VFAC_CTL1_FREEZE_CAPTURE_SYNC);
	} else {
		dp->vfac1 &= ~1;

		cyberpro_grphw8(VFAC_CTL1, dp->vfac1 |
				VFAC_CTL1_FREEZE_CAPTURE|
				VFAC_CTL1_FREEZE_CAPTURE_SYNC);

		cyberpro_grphw8(0xb5, old & ~0xfe);
	}

	/*
	 * wait for things to settle
	 *  (FIXME: do we need to wait so long?)
	 */
	mdelay(10);

	/*
	 * unfreeze capture
	 */
	cyberpro_grphw8(VFAC_CTL1, dp->vfac1);

	/*
	 * Reset the memory sequencer
	 */
	reset_seq();
}

/*
 * Enable or disable the overlay window.  Note
 * that for anything useful to be displayed, we
 * must have capture enabled.
 */
static void
cyberpro_grabber_overlay(struct cyberpro_grab_info *dp, int on)
{
	if (on)
		dp->disp_ctl1 |= VID_DISP_CTL1_ENABLE_VID_WINDOW;
	else
		dp->disp_ctl1 &= ~VID_DISP_CTL1_ENABLE_VID_WINDOW;

	cyberpro_grphw8(VID_DISP_CTL1, dp->disp_ctl1);
}

static void
cyberpro_grabber_set_overlay_window(struct cyberpro_grab_info *dp)
{
	unsigned int xscale, yscale;
	unsigned int xoff, yoff;

	/*
	 * X & Y parameters need to take into account the offset
	 * of the screen from the start of DRAM.  While X is running
	 * this should be +0+0.
	 */
	xoff = yoff = 0;

	xoff += dp->win.x;
	yoff += dp->win.y;

	cyberpro_grphw16(EXT_X_START,        xoff);
	cyberpro_grphw16(EXT_Y_START,        yoff);
	cyberpro_grphw16(EXT_X_END,          xoff + dp->win.width);
	cyberpro_grphw16(EXT_Y_END,          yoff + dp->win.height);
	cyberpro_grphw24(EXT_COLOUR_COMPARE, dp->win.chromakey);

	/*
	 * We need to scale from the capture buffer
	 * to the overlay window
	 */
	xscale = (dp->cap_width  * 4096) / dp->win.width;
	yscale = (dp->cap_height * 4096) / dp->win.height;

	cyberpro_grphw16(EXT_DDA_X_INIT,     0x0800);
	cyberpro_grphw16(EXT_DDA_X_INC,      xscale);
	cyberpro_grphw16(EXT_DDA_Y_INIT,     0x0800);
	cyberpro_grphw16(EXT_DDA_Y_INC,      yscale);

	/*
	 * set interlace mode
	 */
	if (dp->win.flags & VIDEO_WINDOW_INTERLACE) {
		dp->vfac3 |= 2;
		dp->cap_miscctl &= ~CAP_CTL_MISC_ODDEVEN;
	} else {
		dp->vfac3 &= ~2;
		dp->cap_miscctl |= CAP_CTL_MISC_ODDEVEN;
	}

	cyberpro_grphw8(VFAC_CTL3,    dp->vfac3);
	cyberpro_grphw8(CAP_CTL_MISC, dp->cap_miscctl);
}

static void
cyberpro_set_capture_window(struct cyberpro_grab_info *dp)
{
	unsigned int xstart, xend, ystart, yend;
	unsigned int phase, pitch;

	if (dp->stream_8bpp) {
		/* 8-bit capture */
		dp->cap_mode1 |= CAP_MODE1_8BIT;
		xstart = 15;
		xend   = xstart + dp->cap_width * 2;
	} else {
		/* 16-bit capture */
		dp->cap_mode1 &= ~CAP_MODE1_8BIT;
		xstart = 7;
		xend   = xstart + dp->cap_width;
	}

	ystart = 18;
	yend   = ystart + dp->cap_height / 2;

	cyberpro_grphw16(CAP_X_START, xstart);
	cyberpro_grphw16(CAP_X_END, xend + 1);
	cyberpro_grphw16(CAP_Y_START, ystart);
	cyberpro_grphw16(CAP_Y_END, yend + 2);

	cyberpro_grphw16(CAP_DDA_X_INIT, 0x0800);
	cyberpro_grphw16(CAP_DDA_X_INC, 0x1000);
	cyberpro_grphw16(CAP_DDA_Y_INIT, 0x0800);
	cyberpro_grphw16(CAP_DDA_Y_INC, 0x1000);

	cyberpro_grphw8(CAP_PITCH, dp->cap_width >> 2);
	cyberpro_grphw8(CAP_MODE1, dp->cap_mode1);

	/*
	 * We also need to change the parameters for
	 * capture->overlay
	 */
	pitch = (dp->cap_width >> 2) & 0x0fff;
	phase = (dp->cap_width + 3) >> 2;

	cyberpro_grphw16(EXT_SRC_WIDTH, pitch | ((phase << 4) & 0xf000));
	cyberpro_grphw8 (EXT_SRC_WIN_WIDTH, phase);
}

static void
cyberpro_set_format(struct cyberpro_grab_info *dp)
{
	dp->vid_fmt &= ~7;

	switch (dp->pic.palette) {
	case VIDEO_PALETTE_RGB565:	dp->vid_fmt |= VID_CAP_VFC_RGB565;	break;
	case VIDEO_PALETTE_RGB24:	dp->vid_fmt |= VID_CAP_VFC_RGB888_24;	break;
	case VIDEO_PALETTE_RGB32:	dp->vid_fmt |= VID_CAP_VFC_RGB888_32;	break;
	case VIDEO_PALETTE_RGB555:	dp->vid_fmt |= VID_CAP_VFC_RGB555;	break;
	case VIDEO_PALETTE_YUV422:	dp->vid_fmt |= VID_CAP_VFC_YUV422;	break;
	}

	cyberpro_grphw8(VID_CAP_VFC, dp->vid_fmt);
}

/*
 * Calculate and set the address of the capture buffer.  Note we
 * also update the extended memory buffer for the overlay window.
 */
static int
cyberpro_set_buffer(struct cyberpro_grab_info *dp, struct video_buffer *b)
{
#define MAX_SIZE	(700*572*2 + 16384)

	if (b->height <= 0 || b->width <= 0 || b->bytesperline <= 0)
		return -EINVAL;

	if ((b->height * b->bytesperline + MAX_SIZE) >= fb_size)
		return -EINVAL;

	dp->buf.base         = b->base;		/* phys base address of display	*/
	dp->buf.width        = b->width;	/* width of display		*/
	dp->buf.height       = b->height;	/* height of display		*/
	dp->buf.depth        = b->depth;	/* depth of display (8/16/24)	*/
	dp->buf.bytesperline = b->bytesperline;	/* width of display		*/

	dp->cap_mem_offset   = dp->buf.bytesperline * dp->buf.height;
	dp->cap_mem_offset  += 16384;		/* allow a bit of slack		*/
	dp->cap_mem_offset >>= 2;

	cyberpro_grphw24(EXT_MEM_START, dp->cap_mem_offset);
	cyberpro_grphw24(CAP_MEM_START, dp->cap_mem_offset);

	return 0;
}

static void
cyberpro_init(struct cyberpro_grab_info *dp)
{
	cyberpro_grphw8(VID_DISP_CTL1, 0x00);

	dp->vfac1 = VFAC_CTL1_PHILIPS;
	dp->vfac2 = VFAC_CTL2_INVERT_FRAME;
	dp->vfac3 = 0;

	cyberpro_grphw8(VFAC_CTL1, dp->vfac1);
	cyberpro_grphw8(VFAC_CTL3, dp->vfac3);
	cyberpro_grphw8(VFAC_CTL2, dp->vfac2);

	cyberpro_grphw8(REG_BANK, REG_BANK_Y);		/* register bank Y */
	cyberpro_grphw8(Y_TV_CTL, 0x80);		/* reg Y */

	cyberpro_grphw8(REG_BANK,  REG_BANK_K);	/* register bank K */
	/* single channel capture,
	 * vert interpolate,
	 * colorkey,
	 * video data clipped to 0x10-0xF0,
	 * linear uv,
	 * but we are not using x2 so this has no effect
	 */
	cyberpro_grphw8(K_CAP_X2_CTL1,  0x40);	/* reg K */

	/* setup capture */
	/* VLF fifo, no mirrors */
	dp->cap_mode1   = CAP_MODE1_ALTFIFO|
			  CAP_MODE1_CCIR656;

	dp->cap_miscctl = CAP_CTL_MISC_DISPUSED |
			  CAP_CTL_MISC_SYNCTZOR |
			  CAP_CTL_MISC_SYNCTZHIGH;

	cyberpro_grphw8(CAP_NEW_CTL1, 0x3f);	/* disable PIP */
	cyberpro_grphw8(CAP_NEW_CTL2, 0xc0);	/* no IRQs */
	cyberpro_grphw8(CAP_MODE2,    0x50);	/* normal */

	/* setup overlay */
	dp->disp_ctl1 = VID_DISP_CTL1_VINTERPOL_OFF |
			VID_DISP_CTL1_NOCLIP;

	cyberpro_grphw8(VID_DISP_CTL1, dp->disp_ctl1);
	cyberpro_grphw16(VID_FIFO_CTL, 0xf87c);
	cyberpro_grphw16(EXT_FIFO_CTL, 0x1010);
	cyberpro_grphw8(VID_FIFO_CTL1, 0x06);

	cyberpro_set_format(dp);
	cyberpro_set_capture_window(dp);
}

static int
cyberpro_alloc_buffer(struct cyberpro_grab_info *dp)
{
	if (dp->buffer == NULL) {
		dp->buffer = (void *)__get_free_pages(GFP_KERNEL, 8);
		if (!dp->buffer)
			return -ENOMEM;
		memzero(dp->buffer, PAGE_SIZE << 8);
	}
	return 0;
}

static int
cyberpro_grabber_open(struct video_device *dev, int flags)
{
	struct cyberpro_grab_info *dp = (struct cyberpro_grab_info *)dev->priv;
	int one = 1, i, ret;

	if (flags || dp->users)
		return -EBUSY;

	dp->users += 1;

	if (dp->users == 1) {
#ifdef USE_IRQ
		ret = request_irq(IRQ_NETWINDER_VGA, cyberpro_interrupt,
				  SA_SHIRQ, "cyberpro-grabber", dp);
#else
		ret = 0;
#endif
		if (ret) {
			dp->users -= 1;
			return ret;
		}

		/*
		 * Wake up the SAA7111 decoder
		 */
		decoder_sleep(0);

		/*
		 * Initialise the VGA chip
		 */
		cyberpro_init(dp);

		i2c_control_device(&cyberpro_bus, I2C_DRIVERID_VIDEODECODER,
				   DECODER_ENABLE_OUTPUT, &one);

		for (i = 0; i < NR_FRAMES; i++)
			dp->frame[i].status = UNUSED;

		dp->frame_idx = 0;
	}

	MOD_INC_USE_COUNT;

	return 0;
}

static void
cyberpro_grabber_close(struct video_device *dev)
{
	struct cyberpro_grab_info *dp = (struct cyberpro_grab_info *)dev->priv;
	int zero = 0;

	if (dp->users == 1) {
		/*
		 * Disable interrupt
		 */
		dp->vfac3 &= ~VFAC_CTL3_CAP_IRQ;
		cyberpro_grphw8(VFAC_CTL3, dp->vfac3);

		/*
		 * Shut down overlay
		 */
		if (dp->ovl_active)
			cyberpro_grabber_overlay(dp, 0);
		dp->ovl_active = 0;

		/*
		 * Shut down capture
		 */
		if (dp->cap_active)
			cyberpro_grabber_capture(dp, 0);
		dp->cap_active = 0;

		/*
		 * Free any memory
		 */
		if (dp->buffer) {
			free_pages((unsigned long)dp->buffer, BUFFER_ORDER);
			dp->buffer = NULL;
		}

		dp->mmaped = 0;

		/*
		 * Put the SAA7111 decoder back to sleep
		 */
		i2c_control_device(&cyberpro_bus, I2C_DRIVERID_VIDEODECODER,
				   DECODER_ENABLE_OUTPUT, &zero);

		decoder_sleep(1);
#ifdef USE_IRQ
		free_irq(IRQ_NETWINDER_VGA, dp);
#endif
	}

	dp->users -= 1;

	MOD_DEC_USE_COUNT;
}

/*
 * We don't support writing to the grabber
 */
static long
cyberpro_grabber_write(struct video_device *dev, const char *buf, unsigned long count, int noblock)
{
	return -EINVAL;
}

static void
cyberpro_dbg_dump(void)
{
	int i;
	unsigned char idx[] = { 0x30, 0x3e, 0x98, 0x99, 0x9a, 0x9b, 0x9c, 0x9d, 0xa6, 0xa7, 0xa8,
		0xa9, 0xaa, 0xab, 0xac, 0xad };
	for (i = 0; i < sizeof(idx); i++)
		printk("%s%02x%c", (i & 15) ? "" : KERN_DEBUG,
				cyberpro_grphr8(idx[i]),
				(i & 15) == 15 ? '\n' : ' ');
	if ((i & 15) != 0)
		printk("\n");
}

static void cyberpro_interrupt(int nr, void *dev_id, struct pt_regs *regs)
{
	struct cyberpro_grab_info *dp = (struct cyberpro_grab_info *)dev_id;
	unsigned char old_grphidx;
	int idx;

	old_grphidx = cyber2000_inb(0x3ce);

	cyberpro_grphw8(VFAC_CTL3, dp->vfac3);

	if (dp->mmaped) {
		idx = dp->frame_idx;

		printk(KERN_DEBUG "cyberpro: frame %d complete\n", idx);

		/*
		 * If we were grabbing to a buffer, mark
		 * the buffer as complete, and move to the
		 * next buffer.
		 */
		if (dp->frame[idx].status == GRABBING) {
			dp->frame[idx].status = DONE;

			idx += 1;
			if (idx >= NR_FRAMES)
				idx = 0;
		}

		/*
		 * If the next buffer is ready for grabbing,
		 * set up the bus master registers for the
		 * transfer.
		 */
		if (dp->frame[idx].status == GRABBING) {
			printk(KERN_DEBUG "cyberpro: switching to frame %d\n",
			       idx);
		}

		dp->frame_idx = idx;
	}

	wake_up(&dp->waitq);

	cyber2000_outb(old_grphidx, 0x3ce);
}

/*
 * Our general plan here is:
 *  1. Get one frame's worth of physical contiguous memory
 *  2. Set the CyberPro to perform a BM-DMA of one frame to this memory
 *  3. Copy the frame to the userspace
 *  4. Free the physical memory
 * Future enhancements may be to keep the physical memory cached somewhere,
 * but it's of the order of 512KB (a 700x572 grab is 392KB).
 *
 * The Rebel.com original vidcap.o driver freezes the capture, and then
 * copies (maybe with conversion) the data from the VGA capture buffer
 */
static long
cyberpro_grabber_read(struct video_device *dev, char *buf, unsigned long count, int noblock)
{
	struct cyberpro_grab_info *dp = (struct cyberpro_grab_info *)dev->priv;

#if 0
	unsigned long buspage;

	/*
	 * The buffer must not be mmaped
	 * for this to work.
	 */
	if (dp->mmaped)
		return -EINVAL;

	ret = cyberpro_alloc_buffer(dp);
	if (ret)
		return ret;

	if (count > 4096<<8)
		count = 4096<<8;

	buspage = virt_to_bus(dp->buffer);
	printk(KERN_DEBUG "buspage = %08lx\n", buspage);

	/*
	 * Setup the video-side bus master registers
	 */
	cyberpro_grphw24(0xa7, dp->cap_mem_offset);
	cyberpro_grphw8(0xaa,  dp->cap_width >> 2);

	/*
	 * Where do we set the PCI address and legnth
	 * for the bus master transfer?
	 */

	/*
	 * clear flip status
	 */
	cyberpro_grphw8(0xac, 0x20);
	cyberpro_grphw8(0xac, 0x00);

	cyberpro_dbg_dump();
#if 0
	cyberpro_grphw8(REG_BANK, REG_BANK_Y);
	cyberpro_grphw8(0x3e, 0x07);
	cyberpro_grphw8(REG_BANK, REG_BANK_Y);
	cyberpro_grphw8(0xad, 0x00);

	cyberpro_grphw8(0xaa, 0x00);
	cyberpro_grphw8(0xac, 0x00);
	cyberpro_grphw8(0xa6, 0x04);
	cyberpro_grphw8(0xa6, 0x01);
#endif
	/*
	 * Wait for a frame.  This is only testing code, so
	 * lets just guess it, just to see if we actually
	 * capture anything at all.  Really, we need to be
	 * woken from an interrupt routine or similar.
	 */
	set_current_state(TASK_INTERRUPTIBLE);
	schedule_timeout(HZ/50);
#if 0
	cyberpro_grphw8(0xa6, 0x04);
	cyberpro_grphw8(0xa6, 0x00);
	/*
	 * Writing '0' into 0xa7 does not seem to
	 * cause any strange effects on a running capture
	 */
	cyberpro_grphw24(0xa7, 0x00);
	cyberpro_grphw8(0x3e, 0x00);
#endif
	cyberpro_dbg_dump();

	if (copy_to_user(buf, dp->buffer, count))
		count = -EFAULT;
#else
	char *p = fb_base + (dp->cap_mem_offset << 2);
	if (count > 700*572*2)
		count = 700*572*2;

	/*
	 * freeze capture asynchronously
	 *  (FIXME: do we need to set VFAC_CTL1_FREEZE_CAPTURE_SYNC?)
	 */
	cyberpro_grphw8(VFAC_CTL1, dp->vfac1 |
				VFAC_CTL1_FREEZE_CAPTURE|
				VFAC_CTL1_FREEZE_CAPTURE_SYNC);

	if (!(cyberpro_grphr8(0xe5) & 0x02))
		printk(KERN_WARNING
			"%s: video capture freeze unsuccessful\n",
			dp->dev->name);

	/*
	 * Reset the memory sequencer
	 */
	reset_seq();

	if (copy_to_user(buf, p, count))
		count = -EFAULT;

	/*
	 * unfreeze capture
	 */
	cyberpro_grphw8(VFAC_CTL1, dp->vfac1);

	/*
	 * Reset the memory sequencer
	 */
	reset_seq();
#endif

	return count;
}

static int
cyberpro_grabber_ioctl(struct video_device *dev, unsigned int cmd, void *arg)
{
	struct cyberpro_grab_info *dp = (struct cyberpro_grab_info *)dev->priv;

	switch (cmd) {
	case VIDIOCGCAP:
		return copy_to_user(arg, &dp->cap, sizeof(dp->cap))
			? -EFAULT : 0;

	case VIDIOCGCHAN:
		return copy_to_user(arg, &dp->chan, sizeof(dp->chan))
			? -EFAULT : 0;

	case VIDIOCGPICT:
		return copy_to_user(arg, &dp->pic, sizeof(dp->pic))
			? -EINVAL : 0;

	case VIDIOCGWIN:
		return copy_to_user(arg, &dp->win, sizeof(dp->win))
			? -EINVAL : 0;

	case VIDIOCGFBUF:
		return copy_to_user(arg, &dp->buf, sizeof(dp->buf))
			? -EINVAL : 0;

	case VIDIOCGUNIT:
	{
		struct video_unit unit;

		unit.video    = dev->minor;
		unit.vbi      = VIDEO_NO_UNIT;
		unit.radio    = VIDEO_NO_UNIT;
		unit.audio    = VIDEO_NO_UNIT;
		unit.teletext = VIDEO_NO_UNIT;

		return copy_to_user(arg, &unit, sizeof(unit))
			? -EINVAL : 0;
	}

	case VIDIOCSCHAN:
	{
		struct video_decoder_capability vdc;
		struct video_channel v;
		int ok;

		copy_from_user_ret(&v, arg, sizeof(v), -EFAULT);

		if (v.channel != 0)
			return -EINVAL;

		i2c_control_device(&cyberpro_bus, I2C_DRIVERID_VIDEODECODER,
				   DECODER_GET_CAPABILITIES, &vdc);

		switch (v.norm) {
		case VIDEO_MODE_PAL:
			ok = vdc.flags & VIDEO_DECODER_PAL;
			break;
		case VIDEO_MODE_NTSC:
			ok = vdc.flags & VIDEO_DECODER_NTSC;
			break;
		case VIDEO_MODE_AUTO:
			ok = vdc.flags & VIDEO_DECODER_AUTO;
			break;
		default:
			ok = 0;
		}
		if (!ok)
			return -EINVAL;

		dp->chan.channel = v.channel;
		dp->chan.norm    = v.norm;

		i2c_control_device(&cyberpro_bus, I2C_DRIVERID_VIDEODECODER,
				   DECODER_SET_NORM, &dp->chan.norm);

		return 0;
	}

	case VIDIOCSPICT:
	{
		struct video_picture p;
		int update_palette;

		copy_from_user_ret(&p, arg, sizeof(p), -EFAULT);

		update_palette = dp->pic.palette != p.palette;

		/* we can't change the palette while capture is
		 * in operation.
		 */
		if (update_palette && dp->cap_active)
			return -EINVAL;

		switch (p.palette) {
		case VIDEO_PALETTE_YUV422:
			if (p.depth != 8)
				return -EINVAL;
			break;
#if 0 /* can we do these others? */
		case VIDEO_PALETTE_RGB555:
			if (p.depth != 15)
				return -EINVAL;
			break;

		case VIDEO_PALETTE_RGB565:
			if (p.depth != 16)
				return -EINVAL;
			break;

		case VIDEO_PALETTE_RGB24:
			if (p.depth != 24)
				return -EINVAL;
			break;

		case VIDEO_PALETTE_RGB32:
			if (p.depth != 32)
				return -EINVAL;
			break;
#endif
		default:
		 	return -EINVAL;
		}

		dp->pic = p;

		/* p.depth sets the capture depth */
		/* p.palette sets the capture palette */

		i2c_control_device(&cyberpro_bus, I2C_DRIVERID_VIDEODECODER,
				   DECODER_SET_PICTURE, &p);

		if (update_palette)
			cyberpro_set_format(dp);
		return 0;
	}

	case VIDIOCSWIN:	/* set the size & position of the overlay window */
	{
		struct video_window w;
		int diff;

		if (!dp->buf_set)
			return -EINVAL;

		copy_from_user_ret(&w, arg, sizeof(w), -EFAULT);

		if (w.width < dp->cap.minwidth || w.height < dp->cap.minheight ||
		    w.width > dp->cap.maxwidth || w.height > dp->cap.maxheight ||
		    w.clipcount)
			return -EINVAL;

		/*
		 * Bound the overlay window by the size of the screen
		 */
		if (w.x < 0)
			w.x = 0;
		if (w.y < 0)
			w.y = 0;
		if (w.x + w.width > dp->buf.width)
			w.width = dp->buf.width - w.x;
		if (w.y + w.height > dp->buf.height)
			w.height = dp->buf.height - w.y;

		diff = dp->win.x      != w.x      ||
		       dp->win.y      != w.y      ||
		       dp->win.width  != w.width  ||
		       dp->win.height != w.height ||
		       dp->win.chromakey != w.chromakey ||
		       dp->win.flags  != w.flags  ||
		       !dp->win_set;

		if (diff) {
			dp->win.x         = w.x;
			dp->win.y         = w.y;
			dp->win.width     = w.width;
			dp->win.height    = w.height;
			dp->win.chromakey = w.chromakey;
#if 0
			dp->win.flags     = w.flags;
#else
			dp->win.flags     = VIDEO_WINDOW_INTERLACE;
#endif

			if (dp->ovl_active)
				cyberpro_grabber_overlay(dp, 0);

			cyberpro_grabber_set_overlay_window(dp);

			if (dp->ovl_active)
				cyberpro_grabber_overlay(dp, 1);
		}

		dp->win_set = 1;

		return 0;
	}

	case VIDIOCSFBUF:	/* set frame buffer info */
	{
		struct video_buffer b;
		int ret;

		if (!capable(CAP_SYS_ADMIN) || !capable(CAP_SYS_RAWIO))
			return -EPERM;

		if (dp->cap_active)
			return -EINVAL;

		copy_from_user_ret(&b, arg, sizeof(b), -EFAULT);

		ret = cyberpro_set_buffer(dp, &b);
		if (ret == 0) {
			dp->buf_set = 1;
			dp->win_set = 0;
		}

		return ret;
	}

	case VIDIOCCAPTURE:
	{
		int on;

		get_user_ret(on, (int *)arg, -EFAULT);

		if (( on &&  dp->ovl_active) ||
		    (!on && !dp->ovl_active))
			return 0;

		if (on && (!dp->buf_set || !dp->win_set))
			return -EINVAL;

		cyberpro_grabber_capture(dp, on);
		dp->cap_active = on;
		cyberpro_grabber_overlay(dp, on);
		dp->ovl_active = on;

		return 0;
	}

	case VIDIOCSYNC:
	{
		DECLARE_WAITQUEUE(wait, current);
		int buf;

		/*
		 * The buffer must have been mmaped
		 * for this call to work.
		 */
		if (!dp->mmaped)
			return -EINVAL;

		get_user_ret(buf, (int *)arg, -EFAULT);

		if (buf < 0 || buf >= NR_FRAMES)
			return -EINVAL;

		switch (dp->frame[buf].status) {
		case UNUSED:
			return -EINVAL;

		case GRABBING:
			add_wait_queue(&dp->waitq, &wait);
			while (dp->frame[buf].status == GRABBING) {
				set_current_state(TASK_INTERRUPTIBLE);
				schedule();
				if (signal_pending(current))
					return -EINTR;
			}
			remove_wait_queue(&dp->waitq, &wait);
			/*FALLTHROUGH*/
		case DONE:
			dp->frame[buf].status = UNUSED;
			break;
		}
		return 0;
	}

	case VIDIOCMCAPTURE:
	{
		struct video_mmap vmap;

		/*
		 * The buffer must have been mmaped
		 * for this call to work.
		 */
		if (!dp->mmaped)
			return -EINVAL;

		copy_from_user_ret(&vmap, arg, sizeof(vmap), -EFAULT);

		/*
		 * We must capture in YUV422
		 */
		if (vmap.frame >= NR_FRAMES || vmap.format != VIDEO_PALETTE_YUV422 ||
		    vmap.width < dp->cap.minwidth || vmap.height < dp->cap.minheight ||
		    vmap.width > dp->cap.maxwidth || vmap.height > dp->cap.maxheight)
			return -EINVAL;

		if (dp->frame[vmap.frame].status == GRABBING)
			return -EBUSY;

		dp->frame[vmap.frame].width  = vmap.width;
		dp->frame[vmap.frame].height = vmap.height;
		dp->frame[vmap.frame].status = GRABBING;

		dp->vfac3 |= VFAC_CTL3_CAP_IRQ;
		cyberpro_grphw8(VFAC_CTL3, dp->vfac3);
		return 0;
	}

	case VIDIOCGMBUF:
	{
		struct video_mbuf vmb;
		unsigned int frame_size, frame_offset, i;

		frame_size   = dp->cap.maxwidth * dp->cap.maxheight * 2;
		frame_offset = 0;

		for (i = 0; i < NR_FRAMES; i++) {
			vmb.offsets[i] = frame_offset;
			frame_offset  += frame_size;
		}

		vmb.size   = frame_offset;
		vmb.frames = NR_FRAMES;

		return copy_to_user(arg, &vmb, sizeof(vmb)) ? -EFAULT : 0;
	}

	case VIDIOCGCAPTURE:	/* no subcaptures (yet) */
	case VIDIOCSCAPTURE:
	case VIDIOCGTUNER:	/* no tuner */
	case VIDIOCSTUNER:
		return -EINVAL;
	}

	return -EINVAL;
}

static int
cyberpro_grabber_mmap(struct video_device *dev, const char *addr, unsigned long size)
{
	struct cyberpro_grab_info *dp = (struct cyberpro_grab_info *)dev->priv;
	unsigned long vaddr = (unsigned long)addr;
	int ret;

	if (size > PAGE_ALIGN(NR_FRAMES * dp->cap.maxwidth * dp->cap.maxheight * 2))
		return -EINVAL;

	ret = cyberpro_alloc_buffer(dp);
	if (ret == 0)
		ret = remap_page_range(vaddr, virt_to_phys(dp->buffer), size, PAGE_SHARED);
	if (ret == 0)
		dp->mmaped = 1;

	return ret;
}

static int __init
cyberpro_grabber_init_done(struct video_device *dev)
{
	struct cyberpro_grab_info *dp;
	int zero = 0;

	dp = kmalloc(sizeof(*dp), GFP_KERNEL);
	if (!dp)
		return -ENOMEM;

	memset(dp, 0, sizeof(*dp));

	dev->priv = dp;
	dp->dev = dev;

	strcpy(dp->cap.name, dev->name);
	dp->cap.type		= dev->type;
	dp->cap.channels	= 1;
	dp->cap.audios		= 0;
	dp->cap.minwidth	= 700;
	dp->cap.maxwidth	= 700;
	dp->cap.minheight	= 572;
	dp->cap.maxheight	= 572;
	
	strcpy(dp->chan.name, "Composite");
	dp->chan.tuners		= 0;
	dp->chan.flags		= 0;
	dp->chan.type		= VIDEO_TYPE_TV;
	dp->chan.norm		= VIDEO_MODE_AUTO;

	dp->pic.brightness	= 32768;
	dp->pic.hue		= 32768;
	dp->pic.colour		= 32768;
	dp->pic.contrast	= 32768;
	dp->pic.whiteness	= 0;
	dp->pic.depth		= 8;
	dp->pic.palette		= VIDEO_PALETTE_YUV422;

	/* dp->buf is setup by the user		*/
	/* dp->win is setup by the user		*/
	/* dp->cap_mem_offset setup by dp->buf	*/

	dp->stream_8bpp		= 1;
	dp->vid_fmt		= VID_CAP_VFC_YUV422;
	dp->cap_width		= 700;	/* assert((cap_width & 3) == 0) */
	dp->cap_height		= 572;

	init_waitqueue_head(&dp->waitq);

	i2c_control_device(&cyberpro_bus, I2C_DRIVERID_VIDEODECODER,
			   DECODER_SET_NORM, &dp->chan.norm);
	i2c_control_device(&cyberpro_bus, I2C_DRIVERID_VIDEODECODER,
			   DECODER_SET_PICTURE, &dp->pic);
	i2c_control_device(&cyberpro_bus, I2C_DRIVERID_VIDEODECODER,
			   DECODER_ENABLE_OUTPUT, &zero);

	return 0;
}

static struct video_device cyberpro_grabber = {
	"",
	VID_TYPE_CAPTURE |
	VID_TYPE_OVERLAY |
	VID_TYPE_CHROMAKEY,
	0,
	cyberpro_grabber_open,
	cyberpro_grabber_close,
	cyberpro_grabber_read,
	cyberpro_grabber_write,
	NULL,
	cyberpro_grabber_ioctl,
	cyberpro_grabber_mmap,
	cyberpro_grabber_init_done,
	NULL,
	0,
	0
};

int init_cyber2000fb_viddev(struct video_init *unused)
{
	struct cyberpro_info info;
	int ret;

	if (!cyber2000fb_attach(&info))
		return -ENXIO;

	strcpy(cyberpro_grabber.name, info.dev_name);
	strcpy(cyberpro_bus.name, info.dev_name);

	CyberRegs         = info.regs;
	fb_base           = info.fb;
	fb_size           = info.fb_size;
	cyberpro_bus.data = info.regs;
cyber2000_outb(0x3c, 0x3ce);
cyber2000_grphw(0x3c, cyber2000_inb(0x3cf) | 0x80);
cyberpro_dbg_dump();
{ int i, j;
  for (i = 0; i < 256; i += 32) {
    printk("%03X:", i);
    for (j = i; j < i + 32; j+=4) {
      printk(" %08lX", readl(CyberRegs + 0xbc000 + j));
    }
    printk("\n");
  }
}
	/*
	 * wake up the decoder
	 */
	decoder_sleep(0);

	ret = i2c_register_bus(&cyberpro_bus);
	if (ret == 0) {
		ret = video_register_device(&cyberpro_grabber, VFL_TYPE_GRABBER);

		if (ret)
			i2c_unregister_bus(&cyberpro_bus);
	}

	/*
	 * put the decoder back to sleep
	 */
	decoder_sleep(1);

	return ret;
}

#ifdef MODULE
int __init init_module(void)
{
	int ret;

	ret = init_cyber2000fb_viddev(NULL);

	return ret;
}

void cleanup_module(void)
{
	video_unregister_device(&cyberpro_grabber);
	kfree(cyberpro_grabber.priv);
	i2c_unregister_bus(&cyberpro_bus);
	cyber2000fb_detach();
}
#endif
